"""
How to quantum just-in-time (QJIT) compile Grover's algorithm with Catalyst
====================================================================

"""

######################################################################
# `Grover's algorithm </codebook/#05-grovers-algorithm>`__ is an `oracle
# </codebook/04-basic-quantum-algorithms/02-the-magic-8-ball/>`__-based quantum algorithm, first
# proposed by Lov Grover in 1996 [#Grover1996]_, to solve unstructured search problems using a
# `quantum computer <https://pennylane.ai/qml/quantum-computing/>`__. For example, we could use
# Grover's algorithm to search for a phone number in a randomly ordered database containing
# :math:`N` entries and say (with high probability) that the database contains that number by
# performing :math:`O(\sqrt{N})` queries on the database, whereas a classical search algorithm would
# require :math:`O(N)` queries to perform the same task.
#
# More formally, the *unstructured search problem* is defined as a search for a string of bits in a
# list containing :math:`N` items given an *oracle access function* :math:`f(x).` This function is
# defined such that :math:`f(x) = 1` if :math:`x` is the bitstring we are looking for (the
# *solution*), and :math:`f(x) = 0` otherwise. The generalized form of Grover's algorithm accepts
# :math:`M` solutions, with :math:`1 \leq M \leq N.`
#
# In this tutorial, we will implement the generalized Grover's algorithm using `Catalyst
# <https://docs.pennylane.ai/projects/catalyst>`__, a quantum just-in-time (QJIT) compiler framework
# for PennyLane, which makes it possible to compile, optimize, and execute hybrid quantum–classical
# workflows. We will also measure the performance improvement we get from using Catalyst with
# respect to the native Python implementation and show that the runtime performance of circuits
# compiled with Catalyst can be approximately an order of magnitude faster.
# 
# .. figure:: ../_static/demo_thumbnails/opengraph_demo_thumbnails/OGthumbnail_qjit_compile_grovers_algorithm_with_catalyst.png
#     :align: center
#     :width: 70%
#     :target: javascript:void(0)


######################################################################
# Generalized Grover's algorithm with PennyLane
# ---------------------------------------------
#
# In the :doc:`Grover's Algorithm <tutorial_grovers_algorithm>` tutorial, we saw how to implement
# the generalized Grover's algorithm in PennyLane. The procedure is as follows:
#
# #. Initialize the system to an equal superposition over all states.
# #. Perform :math:`r(N, M)` *Grover iterations*:
#
#    #. Apply the unitary *oracle operator*, :math:`U_\omega,` implemented using
#       :class:`~.pennylane.FlipSign`, for each solution index :math:`\omega.`
#    #. Apply the *Grover diffusion operator*, :math:`U_D,` implemented using
#       :class:`~.pennylane.GroverOperator`.
#
# #. Measure the resulting quantum state in the computational basis.
#
# We also saw (in Ref. [#NandC2000]_) that the optimal number of Grover iterations to find the
# solution is given by
#
# .. math:: r(N, M) \approx \left \lceil \frac{\pi}{4} \sqrt{\frac{N}{M}} \right \rceil .
#
# For simplicity, throughout the rest of this tutorial we will consider the search for the :math:`M
# = 2` solution states :math:`\vert 0\rangle ^{\otimes n}` and :math:`\vert 1\rangle ^{\otimes n},`
# where :math:`n = \log_2 N` is the number of qubits, in a "database" of size :math:`N = 2^n`
# containing all possible :math:`n`-qubit states.
#
# First, we'll import the required packages and define the Grover's algorithm circuit, as we did in
# the :doc:`previous tutorial <tutorial_grovers_algorithm>`.






































######################################################################
# We'll begin with a circuit defined using the default state-simulator device, ``"default.qubit"``,
# as our baseline. See the documentation in :func:`~.pennylane.device` for a list of other supported
# devices. To run our performance benchmarks, we'll increase the number of qubits in our circuit to
# :math:`n = 12.`














######################################################################
# Let's quickly confirm that Grover's algorithm correctly identified the solution states
# :math:`\vert 0\rangle ^{\otimes n}` and :math:`\vert 1\rangle ^{\otimes n}` as the most likely
# states to be measured.



















######################################################################
# It worked! We are now ready to QJIT compile our Grover's algorithm circuit.


######################################################################
# Quantum just-in-time compiling the circuit
# ------------------------------------------
#
# Catalyst is developed natively for `PennyLane's high-performance simulators
# <https://pennylane.ai/performance/>`__ and, at the time of writing, does not support the
# ``"default.qubit"`` state-simulator device. Let's first define a new circuit using `Lightning
# <https://docs.pennylane.ai/projects/lightning>`__, which is a PennyLane plugin that provides more
# performant state simulators written in C++. See the :doc:`Catalyst documentation
# <catalyst:dev/devices>` for the full list of devices supported by Catalyst.









######################################################################
# Then, to QJIT compile our circuit with Catalyst, we simply wrap it with :func:`~pennylane.qjit`.




######################################################################
# .. note::
#
#     The Catalyst :class:`~.qjit` decorator supports capturing control flow when specified using
#     the :func:`~pennylane.for_loop`, :func:`~pennylane.while_loop`, and :func:`~pennylane.cond`
#     functions, or additionally, can automatically capture native Python control flow via
#     experimental :doc:`AutoGraph <catalyst:dev/autograph>` support.
#
#     In this tutorial, however, you'll notice that our ``grover_circuit`` function is able to use
#     native Python control flow without the need to convert the Python ``for`` loops to the
#     QJIT-compatible :func:`~.for_loop`, for instance, and without using AutoGraph. The reason we
#     are able to do so here is twofold:
#
#     * The circuit we have compiled, ``circuit_lightning``, is *unparameterized*, meaning it takes
#       in no input arguments. Thus, the control flow of the circuit does not depend on any dynamic
#       variables (whose values are known only at run time).
#
#     * The ranges of the ``for`` loops depend only on static variables (i.e., constants known at
#       compile time), in this case Python-native numerics and lists, and NumPy arrays.
#
#     Hence, the complete control flow of the circuit is known at compile time, which allows us to
#     use native Python control-flow statements.
#
#     See the :doc:`Sharp bits and debugging tips <catalyst:dev/sharp_bits>` section of the Catalyst
#     documentation for more details on this subject.


######################################################################
# We now have our QJIT object, ``circuit_qjit``. A small detail to note in this case is that because
# the function ``circuit_lightning`` takes no input arguments, Catalyst will, in fact,
# *ahead-of-time* (AOT) compile the circuit at instantiation, meaning that when we call this QJIT
# object for the first time, the compilation will have already taken place, and Catalyst will
# execute the compiled code. With JIT compilation, by contrast, the compilation is triggered at the
# first call site rather than at instantiation. See the `Compilation Modes
# <https://docs.pennylane.ai/projects/catalyst/en/stable/dev/quick_start.html#compilation-modes>`__
# documentation in the :doc:`Catalyst Quick Start <catalyst:dev/quick_start>` guide for more
# information on the difference between JIT and AOT compilation.
#
# The compilation step will incur some runtime overhead, which we will measure below. Let's first
# call the compiled circuit and confirm that we get the same results.





######################################################################
# Indeed, we get the same results as before: the compiled circuit has correctly identified the
# solution states :math:`\vert 0\rangle ^{\otimes n}` and :math:`\vert 1\rangle ^{\otimes n}` as the
# most likely states to be measured. We can also compare the results more rigorously by comparing
# element-wise the computed probability of every state (within the given floating-point tolerance):





######################################################################
# Success!


######################################################################
# Benchmarking
# ------------
#
# Let's start profiling the circuits we have defined. We have four function executions in total to
# profile:
#
# #. Executing the circuit using ``"default.qubit"``.
# #. Executing the circuit using ``"lightning.qubit"``.
# #. Compiling the circuit with Catalyst, to measure the AOT compilation overhead.
# #. Calls to the QJIT-compiled circuit, to measure the circuit execution time.
#
# We'll use the `timeit <https://docs.python.org/3/library/timeit.html>`__ module part of the Python
# Standard Library to measure the runtimes. To improve the statistical precision of these
# measurements, we'll repeat the operations for items (2) and (4) five times; item (1) is slow, and
# item (3) is only run once by construction, so we will not repeat these operations.

































































######################################################################
# Let's plot these runtimes as a bar chart to compare them visually.























######################################################################
# This plot illustrates the power of Catalyst: by simply wrapping our Grover's algorithm circuit as
# a QJIT-compiled object (or AOT-compiled in this case) with :func:`~pennylane.qjit`, we have
# achieved execution runtimes approximately an order of magnitude shorter than the PennyLane circuit
# implemented using the ``"lightning.qubit"`` device.
#
# There is one important caveat in this example, however, which is that the compilation step itself
# takes several times longer than the time it takes to run the circuit using the Lightning state
# simulator directly. This is an important factor to consider when deciding whether to QJIT compile
# your own circuits in performance-critical applications. In the case of Grover's algorithm, we only
# needed to execute the circuit once to obtain the solution, so the runtime incurred in the
# compilation step offsets the gain in the compiled circuit's execution time. However, should it be
# necessary to execute your own quantum circuit many times, the runtime savings are compounded with
# every subsequent call to the compiled circuit. Since the compilation step only needs to be
# performed once, the *total* runtime of the QJIT workflow may begin to outperform the baseline
# Lightning workflow as the number of circuit executions increases. [*]_


######################################################################
# Conclusion
# -----------
#
# This tutorial has demonstrated how to just-in-time compile a quantum circuit implementing the
# generalized Grover's algorithm using `Catalyst <https://docs.pennylane.ai/projects/catalyst>`__.
#
# For a circuit with :math:`n = 12` qubits, analogous to a search in a randomly ordered "database"
# containing :math:`N = 2^{12} = 4096` entries, Catalyst offers a circuit-execution runtime
# performance approximately an order of magnitude better than the same circuit implemented using the
# Lightning state-simulator device, with the caveat that the compilation step itself incurs some
# runtime over the workflow with a direct call to the Lightning-implemented circuit.
#
# To learn more about Catalyst and how to use it to compile and optimize your quantum programs and
# workflows, check out the :doc:`Catalyst Quick Start <catalyst:dev/quick_start>` guide.


######################################################################
# References
# ----------
#
# .. [#Grover1996]
#
#     L. K. Grover (1996) "A fast quantum mechanical algorithm for database search". `Proceedings of
#     the Twenty-Eighth Annual ACM Symposium on Theory of Computing. STOC '96. Philadelphia,
#     Pennsylvania, USA: Association for Computing Machinery: 212–219
#     <https://dl.acm.org/doi/10.1145/237814.237866>`__.
#     (arXiv: `9605043 [quant-ph] <https://arxiv.org/abs/quant-ph/9605043>`__)
#
# .. [#NandC2000]
#
#     M. A. Nielsen, and I. L. Chuang (2000) "Quantum Computation and Quantum Information",
#     Cambridge University Press.


######################################################################
# Footnotes
# ---------
#
# .. [*]
#
#     The performance improvements that can be achieved with QJIT compilation will depend on the
#     specific size and topology of your PennyLane circuit.

######################################################################
# About the author
# ----------------